@@ -46,6 +46,9 @@
 #include <gphoto2/gphoto2-port-log.h>
 #include <gphoto2/gphoto2-setting.h>
 
+#include "ptp.h"
+#include "ptp-private.h"
+
 #define PTPIP_VERSION_MAJOR 0x0001
 #define PTPIP_VERSION_MINOR 0x0000
 
@@ -651,10 +654,10 @@ ptp_ptpip_connect (PTPParams* params, const char *address) {
 	}
 	ret = ptp_ptpip_init_command_request (params);
 	if (ret != PTP_RC_OK)
-		return GP_ERROR_IO;
+		return translate_ptp_result (ret);
 	ret = ptp_ptpip_init_command_ack (params);
 	if (ret != PTP_RC_OK)
-		return GP_ERROR_IO;
+		return translate_ptp_result (ret);
 	if (-1 == connect (params->evtfd, (struct sockaddr*)&saddr, sizeof(struct sockaddr_in))) {
 		perror ("connect evt");
 		close (params->cmdfd);
@@ -663,10 +666,10 @@ ptp_ptpip_connect (PTPParams* params, const char *address) {
 	}
 	ret = ptp_ptpip_init_event_request (params);
 	if (ret != PTP_RC_OK)
-		return GP_ERROR_IO;
+		return translate_ptp_result (ret);
 	ret = ptp_ptpip_init_event_ack (params);
 	if (ret != PTP_RC_OK)
-		return GP_ERROR_IO;
+		return translate_ptp_result (ret);
 	gp_log (GP_LOG_DEBUG, "ptpip/connect", "ptpip connected!");
 	return GP_OK;
 }
