@@ -21,7 +21,7 @@
 #include <stdio.h>
 #include <string.h>
 #include <errno.h>
-#include <stdint.h>
+#include <_stdint.h>
 #include <stdlib.h>
 #include <time.h>
 #include <fcntl.h>
@@ -219,7 +219,7 @@ st2205_detect_mem_size(Camera *camera)
 	if (ret)
 		return ret;
 
-	for (i = 0; i < 2; i++) {
+	for (i = 0; i < 3; i++) {
 		ret = st2205_read_block(camera,
 					(524288 / ST2205_BLOCK_SIZE) << i,
 					buf1);
@@ -379,7 +379,6 @@ st2205_add_picture(Camera *camera, int idx, const char *filename,
 {
 	int count;
 	struct image_table_entry entry;
-	struct image_header header;
 
 	count = st2205_read_file_count(camera);
 	if (count < 0) return count;
@@ -407,7 +406,7 @@ st2205_add_picture(Camera *camera, int idx, const char *filename,
 		   seem to care about this, but the windows software does this,
 		   so lets do it too. */
 		memset (&entry, 0, sizeof(entry));
-		entry.address = htole32 (start + size + sizeof(header));
+		entry.address = htole32 (start + size);
 		CHECK (st2205_write_mem (camera, ST2205_FILE_OFFSET (count),
 					 &entry, sizeof(entry)))
 	}
@@ -415,18 +414,29 @@ st2205_add_picture(Camera *camera, int idx, const char *filename,
 	CHECK (st2205_update_fat_checksum (camera))
 	CHECK (st2205_copy_fat (camera))
 
-	memset(&header, 0, sizeof(header));
-	header.marker = ST2205_HEADER_MARKER;
-	header.width  = htobe16(camera->pl->width);
-	header.height = htobe16(camera->pl->height);
-	header.blocks = htobe16((camera->pl->width * camera->pl->height) / 64);
-	header.shuffle_table = shuffle;
-	header.unknown2 = 0x04;
-	header.unknown3 = camera->pl->unknown3[shuffle];
-	header.length = htobe16(size);
-	CHECK (st2205_write_mem (camera, start,
-				 &header, sizeof(header)))
-	CHECK (st2205_write_mem (camera, start + sizeof(header), buf, size))
+	if (camera->pl->compressed) {
+		struct image_header header = { 0, };
+		
+		/* The buffer we are passed in does not include the compressed
+		   image header, but the size does take it into account */
+		size -= sizeof(header);
+
+		/* Create and write compressed image header */
+		header.marker = ST2205_HEADER_MARKER;
+		header.width  = htobe16(camera->pl->width);
+		header.height = htobe16(camera->pl->height);
+		header.blocks = htobe16((camera->pl->width * camera->pl->height) / 64);
+		header.shuffle_table = shuffle;
+		header.unknown2 = 0x04;
+		header.unknown3 = camera->pl->unknown3[shuffle];
+		header.length = htobe16(size);
+		CHECK (st2205_write_mem (camera, start,
+					 &header, sizeof(header)))
+		/* And write the actual image data */
+		CHECK (st2205_write_mem (camera, start + sizeof(header),
+ 					buf, size))
+	} else
+		CHECK (st2205_write_mem (camera, start,	buf, size))
 
 	/* Let the caller know at which index we stored the table entry */
 	return idx;
@@ -511,35 +521,39 @@ st2205_read_raw_file(Camera *camera, int idx, unsigned char **raw)
 
 	GP_DEBUG ("file: %d start at: %08x\n", idx, entry.address);
 
-	CHECK (st2205_read_mem (camera, entry.address,
-				&header, sizeof(header)))
+	if (camera->pl->compressed) {
+		CHECK (st2205_read_mem (camera, entry.address,
+					&header, sizeof(header)))
 
-	if (header.marker != ST2205_HEADER_MARKER) {
-		gp_log (GP_LOG_ERROR, "st2205", "invalid header magic");
-		return GP_ERROR_CORRUPTED_DATA;
-	}
+		if (header.marker != ST2205_HEADER_MARKER) {
+			gp_log (GP_LOG_ERROR, "st2205", "invalid header magic");
+			return GP_ERROR_CORRUPTED_DATA;
+		}
 
-	BE16TOH(header.width);
-	BE16TOH(header.height);
-	BE16TOH(header.length);
-	BE16TOH(header.blocks);
+		BE16TOH(header.width);
+		BE16TOH(header.height);
+		BE16TOH(header.length);
+		BE16TOH(header.blocks);
 
-	if ((header.width != camera->pl->width) ||
-	    (header.height != camera->pl->height)) {
-		gp_log (GP_LOG_ERROR, "st2205",
-			"picture size does not match frame size.");
-		return GP_ERROR_CORRUPTED_DATA;
-	}
+		if ((header.width != camera->pl->width) ||
+		    (header.height != camera->pl->height)) {
+			gp_log (GP_LOG_ERROR, "st2205",
+				"picture size does not match frame size.");
+			return GP_ERROR_CORRUPTED_DATA;
+		}
 
-	if (((header.width / 8) * (header.height / 8)) != header.blocks) {
-		gp_log (GP_LOG_ERROR, "st2205", "invalid block count");
-		return GP_ERROR_CORRUPTED_DATA;
-	}
+		if (((header.width / 8) * (header.height / 8)) != header.blocks) {
+			gp_log (GP_LOG_ERROR, "st2205", "invalid block count");
+			return GP_ERROR_CORRUPTED_DATA;
+		}
 
-	GP_DEBUG ("file: %d header read, size: %dx%d, length: %d bytes\n",
-		  idx, header.width, header.height, header.length);
+		GP_DEBUG ("file: %d header read, size: %dx%d, length: %d bytes\n",
+			  idx, header.width, header.height, header.length);
+
+		size = header.length + sizeof (header);
+	} else
+		size = camera->pl->width * camera->pl->height * 2;
 
-	size = header.length + sizeof (header);
 	*raw = malloc (size);
 	if (!*raw) {
 		gp_log (GP_LOG_ERROR, "st2205", "allocating memory");
@@ -566,10 +580,14 @@ st2205_read_file(Camera *camera, int idx, int **rgb24)
 	CHECK (st2205_read_raw_file (camera, idx, &src))
 
 	header = (struct image_header *) src;
-	ret = st2205_decode_image (camera->pl,
-				   src + sizeof (struct image_header),
-				   be16toh (header->length), rgb24,
-				   header->shuffle_table);
+	if (camera->pl->compressed)
+		ret = st2205_decode_image (camera->pl,
+					   src + sizeof (struct image_header),
+					   be16toh (header->length), rgb24,
+					   header->shuffle_table);
+	else
+		ret = st2205_rgb565_to_rgb24 (camera->pl, src, rgb24);
+
 	free(src);
 
 	return ret;
@@ -585,9 +603,15 @@ st2205_real_write_file(Camera *camera,
 	struct image_header header;
 	int i, start, end, hole_start = 0, hole_idx = 0;
 
-	size = st2205_code_image (camera->pl, rgb24, buf, shuffle,
-				  allow_uv_corr);
-	if (size < 0) return size;
+	if (camera->pl->compressed) {
+		size = st2205_code_image (camera->pl, rgb24, buf, shuffle,
+					  allow_uv_corr);
+		if (size < 0) return size;
+		/* Make sure we have size for both the data and the compressed
+		   image header added by st2205_add_picture */
+		size += sizeof(header);
+	} else
+		size = st2205_rgb24_to_rgb565 (camera->pl, rgb24, buf);
 
 	count = st2205_read_file_count (camera);
 	if (count < 0) return count;
@@ -613,12 +637,16 @@ st2205_real_write_file(Camera *camera,
 
 			start = entry.address;
 			if (entry.present) {
-				CHECK (st2205_read_mem (camera, start,
-							&header,
-							sizeof(header)))
+				if (camera->pl->compressed) {
+					CHECK (st2205_read_mem (camera, start,
+							      &header,
+							      sizeof(header)))
 
-				BE16TOH(header.length);
-				end = start + sizeof(header) + header.length;
+					BE16TOH(header.length);
+					end = start + sizeof(header) + 
+					      header.length;
+				} else
+					end = start + size;
 			}
 		}
 
@@ -629,8 +657,8 @@ st2205_real_write_file(Camera *camera,
 				int hole_size = start - hole_start;
 				GP_DEBUG ("found a hole at: %08x, of %d bytes "
 					  "(need %d)\n", hole_start, hole_size,
-					  (int)(size + sizeof(header)));
-				if (hole_size < (size + sizeof(header))) {
+					  size);
+				if (hole_size < size) {
 					/* Too small, start searching
 					   for the next hole */
 					hole_start = 0;
@@ -656,7 +684,7 @@ st2205_real_write_file(Camera *camera,
 	}
 	
 	/* No freespace found, try again with uv correction tables disabled */
-	if (allow_uv_corr)
+	if (camera->pl->compressed && allow_uv_corr)
 		return st2205_real_write_file (camera, filename, rgb24, buf,
 					       shuffle, 0);
 
@@ -670,7 +698,7 @@ st2205_write_file(Camera *camera,
 	const char *filename, int **rgb24)
 {
 	/* The buffer must be large enough for the worst case scenario */
-	unsigned char buf[(camera->pl->width * camera->pl->height / 64) * 64];
+	unsigned char buf[camera->pl->width * camera->pl->height * 2];
 	int shuffle;
 
 	shuffle = (long long)rand_r(&camera->pl->rand_seed) *
@@ -735,6 +763,10 @@ st2205_set_time_and_date(Camera *camera, struct tm *t)
 {
 	uint8_t *buf = (uint8_t *)camera->pl->buf;
 
+	/* We cannot do this when operating on a dump */
+	if (camera->pl->mem_dump)
+		return GP_OK;
+
 	memset(buf, 0, 512);
 	buf[0] = 6; /* cmd 6 set time */
 	htobe16a (buf + 1, t->tm_year + 1900);
@@ -801,6 +833,7 @@ st2205_init(Camera *camera)
 	uint16_t *lookup_src, *dest;
 	uint8_t *shuffle_src;
 	int x, y, i, j, shuffle_size, checksum, expected_checksum;
+	int is240x320 = 0;
 	const struct {
 		int width, height, no_tables, usable_tables;
 		unsigned char unknown3[8];
@@ -813,8 +846,7 @@ st2205_init(Camera *camera)
 		  { 0xff, 0xff, 0x01, 0x01, 0x01, 0x01, 0x01 },
 		  0x00025800 },
 		{ 120, 160, 7, 7,
-		  /* FIXME unknown3 needs to be verified */
-		  { 0xff, 0xff, 0x01, 0x01, 0x01, 0x01, 0x01 },
+		  { 0xff, 0xff, 0x04, 0x04, 0x04, 0x04, 0x04 },
 		  0x00030570 },
 		{ 96, 64, 7, 7,
 		  { 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00 },
@@ -833,6 +865,7 @@ st2205_init(Camera *camera)
 		  ST2205_V2_PICTURE_START, 1 },
 		{ }
 	};
+	const int uncompressed_firmware_checksums[] = { 0x00ab02fc, 0 };
 
 	GP_DEBUG ("st2205_init called");
 
@@ -844,6 +877,9 @@ st2205_init(Camera *camera)
 		return GP_ERROR_IO;
 	}
 
+	if (camera->pl->width == 240 && camera->pl->height == 320)
+		is240x320 = 1;
+
 	shuffle_size = (camera->pl->width / 8) * (camera->pl->height / 8);
 	if (shuffle_size > ST2205_SHUFFLE_SIZE) {
 		gp_log (GP_LOG_ERROR, "st2205",
@@ -912,6 +948,9 @@ st2205_init(Camera *camera)
 		if (camera->pl->width	== shuffle_info[i].width &&
 				camera->pl->height == shuffle_info[i].height)
 			break;
+		if (is240x320 && shuffle_info[i].width == 120 &&
+				shuffle_info[i].height == 160)
+			break;
 		shuffle_src += (shuffle_info[i].width *
 				shuffle_info[i].height * 2 / 64) *
 			       (shuffle_info[i].no_tables - 2);
@@ -940,6 +979,23 @@ st2205_init(Camera *camera)
 			}
 			checksum += camera->pl->shuffle[j][i].x;
 			checksum += camera->pl->shuffle[j][i].y;
+			if (is240x320) {
+				camera->pl->shuffle[j][i].x *= 2;
+				camera->pl->shuffle[j][i].y *= 2;
+				camera->pl->shuffle[j][i + 1].x =
+					camera->pl->shuffle[j][i].x + 8;
+				camera->pl->shuffle[j][i + 1].y =
+					camera->pl->shuffle[j][i].y;
+				camera->pl->shuffle[j][i + 2].x =
+					camera->pl->shuffle[j][i].x;
+				camera->pl->shuffle[j][i + 2].y =
+					camera->pl->shuffle[j][i].y + 8;
+				camera->pl->shuffle[j][i + 3].x =
+					camera->pl->shuffle[j][i].x + 8;
+				camera->pl->shuffle[j][i + 3].y =
+					camera->pl->shuffle[j][i].y + 8;
+				i += 3;
+			}
 		}
 
 	if (checksum != expected_checksum) {
@@ -952,6 +1008,29 @@ st2205_init(Camera *camera)
 
 	camera->pl->rand_seed = time(NULL);
 
+	/* Some 96x64 models don't use compression, unfortunately I've found
+	   no way to detect if this is the case, so we keep a list of firmware
+	   checksums to identify these. */
+	for (i = camera->pl->mem_size - camera->pl->firmware_size;
+	     i < camera->pl->mem_size; i += ST2205_BLOCK_SIZE)
+		CHECK (st2205_check_block_present (camera,
+						   i / ST2205_BLOCK_SIZE))
+	checksum = 0;
+	for (i = camera->pl->mem_size - camera->pl->firmware_size;
+	     i < camera->pl->mem_size; i++)
+		checksum += (uint8_t)camera->pl->mem[i];
+
+	GP_DEBUG ("firmware checksum: 0x%08x", checksum);
+
+	for (i = 0; uncompressed_firmware_checksums[i]; i++)
+		if (uncompressed_firmware_checksums[i] == checksum)
+			break;
+
+	if (!uncompressed_firmware_checksums[i])
+		camera->pl->compressed = 1;
+	else
+		camera->pl->compressed = 0;
+
 	return GP_OK;
 }
 
